from ordered_set import OrderedSet
from collections import defaultdict, OrderedDict, Counter
from IPython.display import HTML, display
from graphviz import Graph
from copy import deepcopy
from maquinas.exceptions import NoSymbolInMachine


class PartialProduction:
    def __init__(self, head, body, lps=None):
        self.head = head
        # TODO: check for errors in tuplñe
        if isinstance(body, tuple):
            self.l = tuple(body[0])
            self.r = tuple(body[1])
        if not lps:
            self.lps = OrderedDict()
        else:
            self.lps = lps

    def __key(self):
        return (self.head, self.l, self.r)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, PartialProduction):
            return self.__key() == other.__key()
        return False

    def set_pointer(self, p1, p2, label):
        if p2:
            p = (p1, p2)
        else:
            p = (p1,)
        try:
            self.lps[label].add(p)
        except KeyError:
            self.lps[label] = OrderedSet([p])

    def is_partial_production(self):
        return not len(self.r) == 0

    def get_next_symbol(self):
        return self.r[0]

    def get_head(self):
        return self.head

    def move_next_symbol(self):
        return PartialProduction(self.head, (self.l + self.r[:1], self.r[1:]))

    def move_prev_symbol(self):
        return PartialProduction(self.head, (self.l[:-1], self.l[-1:] + self.r))

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if len(self.l) > 0 or len(self.r) > 0:
            l = "".join(self.l)
            r = "".join(self.r)
            return f"{self.head} → {l}●{r}"
        else:
            return f"{self.head}"

    def __format__(self, format):
        return self.__str__()

    def pointers2str(self):
        res = []
        for l, ps in self.lps.items():
            l = f"{l}"
            ps = ", ".join([f"{p}" for p in ps])
            res.append(f"{l}:{ps}")
        return "; ".join(res)

    def get_pointers(self, l):
        try:
            return self.lps[l]
        except IndexError:
            return []
        except KeyError:
            return []

    def reduce_name(self):
        if len(self.r) > 0:
            l = "".join(self.l)
            r = "".join(self.r)
            return f"{self.head} → {l}●{r}"
        else:
            return f"{self.head}"

    def is_terminal(self):
        return len(self.l) == 0 and len(self.r) == 0

    def is_final(self, root):
        return self.head == root and len(self.r) == 0


class Item:
    def __init__(self, pp, ini, fin=None):
        self.pp = pp
        self.ini = ini
        self.fin = fin

    def get_ini(self):
        return self.ini

    def get_fin(self):
        return self.fin

    def get_pproduction(self):
        return self.pp

    def get_head(self):
        return self.pp.get_head()

    def get_next_symbol(self):
        return self.pp.get_next_symbol()

    def get_span(self):
        if self.fin:
            return self.ini, self.fin
        else:
            return None

    def is_partial_production(self):
        return self.pp.is_partial_production()

    def get(self):
        return self.pp, self.ini, self.fin

    def reduce_name(self):
        if not self.fin is None:
            return f"{self.pp.reduce_name()} {self.ini}∷{self.fin}"
        else:
            return f"{self.pp.reduce_name()} {self.ini}"

    def get_pointers(self, label):
        return self.pp.get_pointers(label)

    def add_pointers(self, pointers):
        return self.pp.add_pointers(pointers)

    def pointers2str(self):
        return self.pp.pointers2str()

    def __str__(self):
        if not self.fin is None:
            return f"{self.pp} {self.ini}∷{self.fin}"
        else:
            return f"{self.pp} {self.ini}"

    def __repr__(self):
        return self.__str__()

    def __format__(self, format):
        return self.__str__()

    def __key(self):
        return (self.pp, self.ini, self.fin)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Item):
            return self.__key() == other.__key()
        return False

    def move_next_symbol(self, ini, fin=None):
        return Item(self.pp.move_next_symbol(), ini=ini, fin=fin)

    def index(self):
        return Item(self.pp, self.ini)

    def is_final(self, root):
        return self.pp.is_final(root) and self.ini == 0

    def is_terminal(self):
        return self.pp.is_terminal()

    def set_pointer(self, p1, p2, l):
        self.pp.set_pointer(p1, p2, l)


# Predictor function
def predictor(item, k, productions, working_table):
    pp = item.get_pproduction()
    next_symbol = pp.get_next_symbol()
    for body in productions[next_symbol]:
        pp = PartialProduction(next_symbol, ([], body))
        working_table.add(Item(pp, k))


# Scanner function
def scanner(item, k, tokens, working_table, chart, terminals):
    symbol = item.get_next_symbol()
    origin = item.get_ini()
    token = tokens[k] if k < len(tokens) else None
    if symbol == token and symbol in terminals:
        new_item = item.move_next_symbol(origin)
        if new_item in chart[k + 1]:
            ix = chart[k + 1].index(new_item)
            new_item = chart[k + 1][ix]
        item_ = Item(item.pp, origin, k)
        if len(item.pp.l) > 0:
            new_item.set_pointer(item_, None, k)
        chart[k + 1].add(new_item)
    elif symbol == "ε":
        new_item = item.move_next_symbol(origin)
        if new_item in chart[k]:
            ix = chart[k].index(new_item)
            new_item = chart[k][ix]
        item_ = Item(item.pp, origin, k)
        working_table.add(new_item)
        if len(item.pp.l) > 0:
            new_item.set_pointer(item_, None, k)


def completer(item, k, working_table, chart):
    origin = item.get_ini()
    head = item.get_head()
    for prev_item in chart[origin]:
        try:
            symbol_canditate = prev_item.get_next_symbol()
        except IndexError:
            continue
        if symbol_canditate == head:
            new_item = prev_item.move_next_symbol(prev_item.ini)
            if new_item in chart[k]:
                ix = chart[k].index(new_item)
                new_item = chart[k][ix]
            elif new_item in working_table:
                ix = working_table.index(new_item)
                new_item = working_table[ix]
            item_ = Item(item.pp, origin, k)
            prev_item_ = Item(prev_item.pp, prev_item.ini, origin)
            if len(prev_item.pp.l) > 0:
                new_item.set_pointer(item_, prev_item_, origin)
            else:
                new_item.set_pointer(item_, None, origin)
            working_table.add(new_item)


def parse(tokens, productions, root, terminals):
    chart = [OrderedSet() for _ in range(len(tokens) + 1)]
    for rule in productions[root]:
        starting_production = PartialProduction(root, ([], rule))
        chart[0].add(Item(starting_production, 0))
    pointers = defaultdict(set)
    for k in range(len(tokens) + 1):
        working_table = OrderedSet(chart[k])
        chart[k].clear()
        # TODO: record the operation
        while len(working_table) > 0:
            item = working_table.pop()
            if item in chart[k]:
                if (len(item.pp.r) > 0 and item.pp.r[0] == "ε") or (
                    len(item.pp.l) > 0 and item.pp.l[0] == "ε"
                ):
                    pass
                else:
                    ix = chart[k].index(item)
                    item_ = chart[k][ix]
                    for l, ps in item.pp.lps.items():
                        item_.pp.lps[l].update(ps)
                    continue
            else:
                chart[k].add(item)

            if item.is_partial_production():
                symbol = item.get_next_symbol()
                if not symbol in terminals and not symbol == "ε":
                    predictor(item, k, productions, working_table)
                else:
                    scanner(item, k, tokens, working_table, chart, terminals)
            else:
                completer(item, k, working_table, chart)
    return chart


def print_chart(string, chart, pointers=False, lspace=40):
    for i, items in enumerate(chart):
        print(f"S({i}):", "{}•{}".format(string[0:i], string[i:]))
        pointers_str = ""
        for j, item in enumerate(items):
            pp = str(item.pp)
            if pointers:
                pointers_str = item.pointers2str()
            print(f"{j:3d} | {pp:20s} |  {item.ini:3d} | {pointers_str}")


def chart2table(string, chart, pointers=False):
    if pointers:
        tbl = [
            "<table><tr><th>No.</th><th>Production</th><th>Origin</th><th>Pointers</th></tr>"
        ]
    else:
        tbl = ["<table><tr><th>No.</th><th>Production</th><th>Origin</th></tr>"]
    for i, items in enumerate(chart):
        tbl.append('<tr><th style="text-align:center" colspan="3">')
        tbl.append(f"S({i}):")
        tbl.append(f"{string[:i]}•{string[i:]}")
        tbl.append("</th></tr>")
        for j, item in enumerate(items):
            pp = str(item.pp)
            tbl.append("<tr>")
            tbl.append(f"<td>{j:3d}</td>")
            tbl.append(f"<td style='text-align:left'>{pp}</td>")
            tbl.append(f"<td>{item.ini:3d}</td>")
            if pointers:
                pointers_str = item.pointers2str()
                tbl.append(f"<td>{pointers_str}</td>")
            tbl.append("</td>")
    tbl.append("</table>")
    return display(HTML("".join(tbl)))


def item2label(item, span=False, symbols={}):
    label = f"{item[0]}"
    label = symbols.get(label, label)
    if span:
        label = f"<{label} <FONT POINT-SIZE='7'>{item[1]}:{item[2]}</FONT>>"
    return label


def _extract_trees(
    stack,
    nodes_dict,
    terminals,
    ancesters=Counter(),
    max_depth=None,
    max_ancesters=None,
    tokens=0,
):
    nchars = 0
    depth = 0
    id = 1
    while len(stack) > 0:
        root, current, children, partials, nchars, depth, ancesters = stack.pop(0)
        if max_depth and depth >= max_depth:
            # yield root
            continue
        prev, currentlist = current
        if prev:
            ancesters[prev["label"]] += 1
        if len(children) == 0:
            if (prev["label"][0] in terminals or prev["label"][0] == "ε") and sum(
                [len(p) for p in partials]
            ) == 0:
                yield root
            else:
                _update_stack(
                    stack,
                    root,
                    partials,
                    nchars,
                    terminals,
                    prev=prev,
                    depth=depth,
                    ancesters=ancesters,
                )
        elif len(children) == 1:
            c = nodes_dict[children[0]]
            u = (
                {
                    "label": c["label"],
                    "id": c["id"],
                    "partial": c["partial"],
                },
                [],
            )
            id += 1
            if max_ancesters and max_ancesters <= ancesters[c["label"]]:
                continue
            currentlist.append(u)
            partials[0].insert(0, (u, c["children"]))
            _update_stack(
                stack,
                root,
                partials,
                nchars,
                terminals,
                prev=prev,
                depth=depth,
                ancesters=ancesters,
            )
        elif len(children) == 2:
            c1 = nodes_dict[children[0]]
            u1 = (
                {
                    "label": c1["label"],
                    "id": c1["id"],
                    "partial": c1["partial"],
                },
                [],
            )
            id += 1
            c2 = nodes_dict[children[1]]
            u2 = (
                {
                    "label": c2["label"],
                    "id": c2["id"],
                    "partial": c2["partial"],
                },
                [],
            )
            id += 1
            if max_ancesters and (
                max_ancesters <= ancesters[c1["label"]]
                or max_ancesters <= ancesters[c2["label"]]
            ):
                continue
            currentlist.append(u1)
            currentlist.append(u2)
            partials[0].insert(0, (u1, c1["children"]))
            partials[0].insert(1, (u2, c2["children"]))
            _update_stack(
                stack,
                root,
                partials,
                nchars,
                terminals,
                prev=prev,
                depth=depth,
                ancesters=ancesters,
            )
    return stack


def _deepcopy(tree, label, link=None):
    p, children = tree
    if "current" in p and p["current"]:
        link = (p, [c for c in children])
        return link, link
    children_ = []
    for c in children:
        c_, link_ = _deepcopy(c, label, link=link)
        children_.append(c_)
        if link_:
            link = link_
    return (p, children_), link


def _update_stack(
    stack, root, partials, nchars, terminals, prev=None, depth=0, ancesters=Counter()
):
    if len(partials) == 0:
        if prev:
            if prev["label"][1] == prev["label"][2]:
                stack.insert(0, (root, (None, []), [], [], nchars, depth, ancesters))
            else:
                stack.insert(
                    0, (root, (None, []), [], [], nchars + 1, depth + 1, ancesters)
                )
        return
    while len(partials) > 0 and len(partials[0]) == 0:
        partials.pop(0)
    if len(partials) == 0:
        return
    current, branches = partials[0].pop(0)
    current[0]["current"] = True
    if len(branches) == 0:
        if current[0]["label"][0] in terminals:
            stack.insert(
                0, (root, current, [], partials, nchars + 1, depth + 1, ancesters)
            )
        elif current[0]["label"][0] == "ε":
            stack.insert(0, (root, current, [], partials, nchars, depth + 1, ancesters))
        else:
            raise NoSymbolInMachine(current[0]["label"][0], terminals)
    if len(branches) > 0:
        root_, current_ = _deepcopy(root, current[0])
        if current_ is None:
            partials.append((current, branches))
            current[0]["current"] = False
        else:
            stack.append(
                (root, current, branches[0], partials, nchars, depth + 1, ancesters)
            )
            for ix in range(1, len(branches)):
                root_, current_ = _deepcopy(root, current[0])
                partials_ = [deepcopy(p) for p in partials]
                stack.append(
                    (
                        root_,
                        current_,
                        branches[ix],
                        partials_,
                        int(nchars),
                        depth + 1,
                        Counter(ancesters),
                    )
                )
    current[0]["current"] = False


def extract_trees(forest, terminals, max_depth=None, max_ancesters=None):
    u0, forest = forest
    u0_ = {
        "label": u0["label"],
        "id": u0["id"],
        "partial": u0["partial"],
    }
    root = (u0_, [])
    _, currentlist = root
    stack = []
    _update_stack(
        stack, root, [[(root, u0["children"])]], 0, terminals, ancesters=Counter()
    )
    return _extract_trees(
        stack,
        forest,
        terminals,
        max_depth=max_depth,
        max_ancesters=max_ancesters,
        tokens=u0["tokens"],
    )


def _graph_tree(
    f,
    tree,
    parent,
    c=0,
    ntree=0,
    symbols={},
    span=False,
    full=False,
    show_id=False,
    **params,
):
    tree = list(tree)
    if len(tree) > 1:
        u, children = tree
    else:
        (u,) = tree
        children = ()
    label = item2label(u["label"], span=span, symbols=symbols)
    u_n = f"u_{c}_{ntree}"
    if show_id:
        params[
            "xlabel"
        ] = f"<<I><FONT POINT-SIZE='8'>u</FONT><FONT POINT-SIZE='5'>{u['id']+1}</FONT></I>>"
    if not u["partial"] or full:
        f.node(u_n, label=label, **params)
        if parent:
            f.edge(parent, u_n)
    else:
        u_n = parent

    for child in children:
        c += 1
        c = _graph_tree(
            f,
            child,
            u_n,
            c=c,
            ntree=ntree,
            symbols=symbols,
            span=span,
            full=full,
            show_id=show_id,
            **params,
        )
    return c


def graph_tree(
    tree,
    symbols={},
    span=False,
    full=False,
    show_id=False,
    node_params={"style": "rounded", "shape": "plain"},
    **params,
):
    params["format"] = "svg"
    params["strict"] = True
    f = Graph(comment="tree", **params)
    c = 0
    _graph_tree(
        f,
        tree,
        None,
        c=c,
        symbols=symbols,
        span=span,
        full=full,
        show_id=show_id,
        **node_params,
    )
    return f


def graph_trees(
    trees,
    symbols={},
    span=False,
    full=False,
    show_id=False,
    node_params={"style": "rounded", "shape": "plain"},
    **params,
):
    params["format"] = "svg"
    params["strict"] = True
    f = Graph(comment="tree", **params)
    for i, tree in enumerate(trees):
        with f.subgraph(name=f"Tree {i}") as f_:
            _graph_tree(
                f_,
                tree,
                None,
                ntree=i,
                symbols=symbols,
                span=span,
                full=full,
                show_id=show_id,
                **node_params,
            )
    return f


def _graph_forest(
    f,
    u,
    forest,
    parent,
    ancesters,
    processed,
    symbols={},
    span=False,
    full=True,
    show_id=False,
    **params,
):
    if not u:
        return

    n = tuple(forest.keys()).index(u["label"])
    u_n = f"u_{u['id']+1}"
    ancesters.add(u["id"])

    node_name = u_n
    original_node_name = u_n
    label = item2label(u["label"], span=span, symbols=symbols)

    if show_id:
        params[
            "xlabel"
        ] = f"<<I><FONT POINT-SIZE='8'>u</FONT><FONT POINT-SIZE='5'>{u['id']+1}</FONT></I>>"
    if not u["partial"] or full:
        f.node(u_n, label=label, **params)
        if parent:
            if not (parent, original_node_name) in processed:
                f.edge(parent, original_node_name, dir="forward")
                processed.add((parent, original_node_name))
    else:
        node_name = parent

    for j, u_child_names in enumerate(u["children"]):
        if len(u["children"]) > 1:
            fake_parent = f"fp_{j}_{original_node_name}"
            f.node(fake_parent, label="", shape="point")
            if not (original_node_name, fake_parent) in processed:
                f.edge(original_node_name, fake_parent)
                processed.add((original_node_name, fake_parent))
            node_name = fake_parent
        for u_child_name in u_child_names:
            u_child = forest[u_child_name]
            if u_child["id"] in ancesters:
                u_n_child = f"u_{u_child['id']+1}"
                if not (node_name, u_n_child) in processed:
                    f.edge(node_name, u_n_child, dir="forward")
                    processed.add((node_name, u_n_child))
                continue
            else:
                _graph_forest(
                    f,
                    u_child,
                    forest,
                    node_name,
                    set(ancesters),
                    processed,
                    symbols=symbols,
                    span=span,
                    full=full,
                    show_id=show_id,
                    **params,
                )


def graph_forest(
    forest,
    symbols={},
    span=False,
    full=True,
    show_id=False,
    node_params={"style": "rounded", "shape": "rectangle"},
    **params,
):
    params["format"] = "svg"
    # params['strict']=True
    f = Graph(comment="tree", **params)
    u0, forest = forest
    _graph_forest(
        f,
        u0,
        forest,
        None,
        set(),
        set(),
        symbols=symbols,
        span=span,
        full=full,
        show_id=show_id,
        **node_params,
    )
    return f


def _make_node(label, elements):
    partial = False if isinstance(label[0], str) else True
    return {
        "label": label,
        "id": len(elements),
        "partial": partial,
        "children": OrderedSet(),
    }  # families


def _extract_forest(u, item, elements, processed, chart, terminals, i_table):
    if len(item.pp.l) == 0:
        return None

    symbol = item.pp.l[-1]
    pos = len(item.pp.l[:-1])
    first = len(item.pp.l) == 1

    if symbol == "ε":
        v_name = (item.pp.head, i_table, i_table)
        v = _make_node(v_name, elements)
        if not v_name in elements:
            elements[v_name] = v
        e_name = ("ε", i_table, i_table)
        e = _make_node(e_name, elements)
        if not e_name in elements:
            elements[e_name] = e
        u["children"].add((e_name,))
        processed.add(item)
        return

    if symbol in terminals and first:
        v_name = (symbol, i_table - 1, i_table)
        v = _make_node(v_name, elements)
        if not v_name in elements:
            elements[v_name] = v
        u["children"].add((v_name,))

    if (not symbol in terminals) and first:
        j = item.ini
        v_name = (symbol, j, i_table)
        v = _make_node(v_name, elements)
        if not v_name in elements:
            elements[v_name] = v
        u["children"].add((v_name,))
        for (q,) in item.pp.lps[j]:
            if not q in processed:
                _extract_forest(
                    v, q, elements, processed, chart, terminals, i_table=i_table
                )

    if symbol in terminals and not first:
        j = item.ini
        v_name = (symbol, i_table - 1, i_table)
        v = _make_node(v_name, elements)
        if not v_name in elements:
            elements[v_name] = v
        w_name = (item.pp.move_prev_symbol(), j, i_table - 1)
        w = _make_node(w_name, elements)
        if not w_name in elements:
            elements[w_name] = w
            for (p_,) in item.pp.lps[i_table - 1]:
                _extract_forest(
                    w, p_, elements, processed, chart, terminals, i_table=i_table - 1
                )
        u["children"].add((w_name, v_name))

    if (not symbol in terminals) and not first:
        j = item.ini
        for l, qps in item.pp.lps.items():
            for ii, (q, p_) in enumerate(qps):
                w_name = (item.pp.move_prev_symbol(), j, l)
                w = _make_node(w_name, elements)
                if not w_name in elements:
                    elements[w_name] = w
                else:
                    w = elements[w_name]
                if not p_ in processed:
                    processed.add(p_)
                    _extract_forest(
                        w, p_, elements, processed, chart, terminals, i_table=l
                    )
                v_name = (symbol, l, i_table)
                v = _make_node(v_name, elements)
                if not v_name in elements:
                    elements[v_name] = v
                else:
                    v = elements[v_name]
                if not q in processed:
                    processed.add(q)
                    _extract_forest(
                        v, q, elements, processed, chart, terminals, i_table=i_table
                    )
                u["children"].add((w_name, v_name))


def extract_forest(root, chart, terminals, tokens):
    elements = OrderedDict()
    processed = OrderedSet()
    i_table = len(chart) - 1
    root_name = (root, 0, i_table)
    u0 = _make_node(root_name, elements)
    elements[u0["label"]] = u0
    u0["tokens"] = tokens
    for item in chart[-1]:
        if item.is_final(root):
            _extract_forest(u0, item, elements, processed, chart, terminals, i_table)
    return u0, elements


def verify_chart(chart, tokens, root):
    if not (len(tokens) + 1) == len(chart):
        return False
    return [
        Item(item.pp, item.ini, len(tokens))
        for item in chart[-1]
        if item.is_final(root)
    ]



# Function to load machines
import re
from maquinas.regular.dfa import *
from maquinas.regular.ndfa import *
from maquinas.regular.ndfa_e import *
from maquinas.contextfree.pda import *
from maquinas.recursivelyenumerable.tspda import *
from maquinas.recursivelyenumerable.tm import *
from maquinas.exceptions import *

re_state = re.compile(r"^\s*(?P<initial>->)?\s*(?P<state>.*[^\]])[^\]]*(?P<final>])?$")
re_z_slash_final = re.compile(r"^s*(?P<Z>[^\/]+)/(?P<slash>.*)(->|→)(?P<final>[^,]+)")
re_z_slash = re.compile(r"^s*(?P<Z>[^\/]+)/(?P<slash>.*)")
re_a_a_final_dir = re.compile(r"^s*(?P<a>[^\/]+)/(?P<aa>.*)(->|→)(?P<final>[^,]+):(?P<Dir>R|L)")


def load_fa(string):
    """Loads a finite automata: DFA, NDFA, NDFA-e from a string

    :param string: recieves a string with the definition of the FA
    :return: An auomata"""

    def _read_row(line):
        """Reads a line of a FA"""
        row = [a.strip() for a in line.split("|")]
        origin = _read_state(row[0])
        return origin, _read_states(row[1:])

    def _read_states(states):
        """Reads a destination states of a FA"""
        return [[s.strip() for s in state.split(",")] for state in states]

    header = False
    delta = {}
    for line in string.split("\n"):
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        if not header:
            sigma = _read_sigma(line)
            header = True
        else:
            origin, states = _read_row(line)
            delta[origin] = states
    type_machine = 1
    if "epsilon" in sigma:
        type_machine = 3
    if "ε" in sigma:
        type_machine = 3
    if type_machine == 1:
        for o, states in delta.items():
            for s in states:
                if len(s) > 1:
                    type_machine = 2
                    break
            if type_machine == 2:
                break
    m = _create_fa_by_type(type_machine)
    for a in sigma:
        try:
            m.add_symbol(a)
        except AlreadyExistsSymbol:
            pass
    A = set()
    for (ini, fin, q_i), states in delta.items():
        if fin:
            A.add(q_i)
        for a, state in zip(sigma, states):
            state = [s for s in state if len(s) > 0]
            if len(state) > 0:
                if type_machine == 1 and len(state) == 1:
                    m.add_transition(q_i, a, state[0], force=True)
                else:
                    m.add_transition(q_i, a, state, force=True)
        if ini:
            m.set_initial_state(q_i)
    m.set_aceptors(A)
    return m

def load_tm(string):
    """Loads a Turing Machine

    :param string: recieves a string with the definition of the PDA
    :return: A PDA"""
    m = TuringMachine()

    def _read_row(line):
        """Reads a line of a PDA"""
        row = [a.strip() for a in line.split("|")]
        origin = _read_state(row[0])
        return origin, _read_states(row[1:])

    def _read_states(states):
        """Reads a destination states of a FA"""
        return [
            [_read_a_a_final_dir(s.strip()) for s in state.split(",") if len(s) > 0]
            for state in states
        ]

    def _read_a_a_final_dir(z_slash_final):
        m_ = re_a_a_final_dir.match(z_slash_final)
        return m_.group("a"), m_.group("aa"), m_.group("final"),m_.group('Dir')

    header = False
    delta = {}
    for line in string.split("\n"):
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        if not header:
            sigma = _read_sigma(line)
            header = True
        else:
            origin, states = _read_row(line)
            delta[origin] = states
    for a in sigma:
        try:
            m.add_symbol(a)
        except AlreadyExistsSymbol:
            pass
    A = set()
    for (ini, fin, q_i), states in delta.items():
        if fin:
            A.add(q_i)
        for a, state in zip(sigma, states):
            for a,a_,q_f,Dir in state:
                m.add_transition(q_i, a, [(q_f, a_, m._ndir(Dir))], force=True, update=True)
        if ini:
            m.set_initial_state(q_i)
    m.set_aceptors(A)
    return m


def load_pda(string):
    """Loads a Push Down Automataton

    :param string: recieves a string with the definition of the PDA
    :return: A PDA"""
    m = PushDownAutomaton()

    def _read_row(line):
        """Reads a line of a PDA"""
        row = [a.strip() for a in line.split("|")]
        origin = _read_state(row[0])
        return origin, _read_states(row[1:])

    def _read_states(states):
        """Reads a destination states of a FA"""
        return [
            [_read_z_slash_final(s.strip()) for s in state.split(",") if len(s) > 0]
            for state in states
        ]

    def _read_z_slash_final(z_slash_final):
        m_ = re_z_slash_final.match(z_slash_final)
        return m._filter(m_.group("Z")), m.tokens(m_.group("slash")), m_.group("final")

    header = False
    delta = {}
    for line in string.split("\n"):
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        if not header:
            sigma = _read_sigma(line)
            header = True
        else:
            origin, states = _read_row(line)
            delta[origin] = states
    for a in sigma:
        try:
            m.add_symbol(a)
        except AlreadyExistsSymbol:
            pass
    A = set()
    for (ini, fin, q_i), states in delta.items():
        if fin:
            A.add(q_i)
        for a, state in zip(sigma, states):
            for z, z_, q_f in state:
                m.add_transition(q_i, a, z, [(q_f, z_)], force=True, update=True)
        if ini:
            m.set_initial_state(q_i)
    m.set_aceptors(A)
    return m


def load_tspda(string):
    """Loads a Two Stack Push Down Automataton

    :param string: recieves a string with the definition of the TSPDA
    :return: A TSPDA"""
    m = TwoStackPushDownAutomaton()

    def _read_row(line):
        """Reads a line of a PDA"""
        row = [a.strip() for a in line.split("|")]
        origin = _read_state(row[0])
        return origin, _read_states(row[1:])

    def _read_states(states):
        """Reads a destination states of a FA"""
        return [
            [
                _read_z_slash_final(z1.strip(), z2_q.strip())
                for z1, z2_q in zip(segs[::2], segs[1::2])
                if len(segs) > 0
            ]
            for state in states
            if (segs := state.split(","))
        ]

    def _read_z_slash_final(z1_slash, z2_slash_final):
        m2_ = re_z_slash_final.match(z2_slash_final)
        m1_ = re_z_slash.match(z1_slash)
        return (
            m._filter(m1_.group("Z")),
            m.tokens(m1_.group("slash")),
            m._filter(m2_.group("Z")),
            m.tokens(m2_.group("slash")),
            m2_.group("final"),
        )

    header = False
    delta = {}
    for line in string.split("\n"):
        line = line.strip()
        if len(line) == 0 or line.startswith("#"):
            continue
        if not header:
            sigma = _read_sigma(line)
            header = True
        else:
            origin, states = _read_row(line)
            delta[origin] = states
    for a in sigma:
        try:
            m.add_symbol(a)
        except AlreadyExistsSymbol:
            pass
    A = set()
    for (ini, fin, q_i), states in delta.items():
        if fin:
            A.add(q_i)
        for a, state in zip(sigma, states):
            for z1, z1_, z2, z2_, q_f in state:
                m.add_transition(
                    q_i, a, z1, z2, [(q_f, (z1_, z2_))], force=True, update=True
                )
        if ini:
            m.set_initial_state(q_i)
    m.set_aceptors(A)
    return m


def _read_state(state):
    """Reads a source state of a FA"""
    m = re_state.match(state)
    return (
        m.group("initial") != None,
        m.group("final") != None,
        m.group("state").strip(),
    )


def _create_fa_by_type(type_machine):
    if type_machine == 1:
        return DeterministicFiniteAutomaton()
    elif type_machine == 2:
        return NonDeterministicFiniteAutomaton()
    elif type_machine == 3:
        return NonDeterministicFiniteAutomaton_epsilon()


def _read_sigma(line):
    """Reads an alphabet sigma"""
    return [a.strip() for a in line.split("|") if len(a.strip()) > 0]


def load_jflap(string, state_prefix=lambda x: f"{x}"):
    """Reads a jflap string file definition

    :param string: string with the content of a .jff JFLAP format
    :param state_prefix: function that attach a prefix to the states. Useful when states are numbersi
    :return: An automata"""
    import xml.etree.ElementTree as ET

    root = ET.fromstring(string)
    type_machine = root.find("type").text

    # TODO load other type of inputs: pda, re, grammar
    if type_machine == "fa":
        m = _load_jflap_fa(root, state_prefix=state_prefix)
    elif type_machine == "pda":
        m = _load_jflap_pda(root, state_prefix=state_prefix)
    if type_machine.startswith("turing"):
        m = _load_jflap_tm(root, state_prefix=state_prefix)
    return m

def _load_jflap_tm(root, state_prefix=lambda x: f"{x}"):
    """Loads a jflap xml TM description"""
    m = TuringMachine()
    A = set()
    id2name = {}
    for e in root.iter("state"):
        if "name" in e.attrib:
            q = state_prefix(e.attrib["name"])
            id2name[e.attrib["id"]] = q
        else:
            q = state_prefix(e.attrib["id"])
            id2name[e.attrib["id"]] = q
        try:
            m.add_state(q)
        except AlreadyExistsState:
            pass
        if not e.find("initial") is None:
            m.set_initial_state(q)
        if not e.find("final") is None:
            A.add(q)
    for e in root.iter("block"):
        if "name" in e.attrib:
            q = state_prefix(e.attrib["name"])
            id2name[e.attrib["id"]] = q
        else:
            q = state_prefix(e.attrib["id"])
            id2name[e.attrib["id"]] = q
        try:
            m.add_state(q)
        except AlreadyExistsState:
            pass
        if not e.find("initial") is None:
            m.set_initial_state(q)
        if not e.find("final") is None:
            A.add(q)

    m.set_aceptors(A)
    for e in root.iter("transition"):
        a = e.find("read").text
        if not a:
            a = m.B
        else:
            try:
                m.add_symbol(a)
            except AlreadyExistsSymbol:
                pass
            try:
                m.add_tsymbol(a)
            except AlreadyExistsSymbol:
                pass
        a_ = e.find("write").text
        if not a_:
            a_ = m.B
        else:
            try:
                m.add_symbol(a_)
            except AlreadyExistsSymbol:
                pass
            try:
                m.add_tsymbol(a_)
            except AlreadyExistsSymbol:
                pass
        Dir = e.find("move").text
        if Dir == 'L':
            Dir = -1
        elif Dir == 'R':
            Dir = 1
        else:
            Dir = 0

        q_i = id2name[e.find("from").text]
        q_f = id2name[e.find("to").text]
        m.add_transition(q_i, a, [(q_f, a_, Dir)], force=False, update=True)

    return m


def _load_jflap_pda(root, state_prefix=lambda x: f"{x}"):
    """Loads a jflap xml PDA description"""
    m = PushDownAutomaton(Z_0="Z")
    A = set()
    id2name = {}
    for e in root.iter("state"):
        if "name" in e.attrib:
            q = state_prefix(e.attrib["name"])
            id2name[e.attrib["id"]] = q
        else:
            q = state_prefix(e.attrib["id"])
            id2name[e.attrib["id"]] = q
        try:
            m.add_state(q)
        except AlreadyExistsState:
            pass
        if not e.find("initial") is None:
            m.set_initial_state(q)
        if not e.find("final") is None:
            A.add(q)
    m.set_aceptors(A)
    for e in root.iter("transition"):
        a = e.find("read").text
        if not a:
            a = "epsilon"
        else:
            try:
                m.add_symbol(a)
            except AlreadyExistsSymbol:
                pass
        q_i = id2name[e.find("from").text]
        q_f = id2name[e.find("to").text]
        pop = e.find("pop").text
        if not pop:
            pop = "epsilon"
        try:
            m.add_qsymbol(a)
        except AlreadyExistsSymbol:
            pass
        push = e.find("push").text
        if not push:
            push = ["epsilon"]
        else:
            push = m.tokens(push)
            for p in push:
                try:
                    m.add_qsymbol(p)
                except AlreadyExistsSymbol:
                    pass
        m.add_transition(q_i, a, pop, [(q_f, push)], force=False)

    return m


def _load_jflap_fa(root, state_prefix=lambda x: f"{x}"):
    """Loads a jflap xml fa description"""

    type_machine = 0
    if any(t.find("read").text == None for t in root.iter("transition")):
        type_machine = 3
    else:
        trans = set()
        for e in root.iter("transition"):
            if (e.find("from").text, e.find("read").text) in trans:
                type_machine = 2
                break
            trans.add((e.find("from").text, e.find("read").text))
            print(trans)
        if type_machine == 0:
            type_machine = 1
    m = _create_fa_by_type(type_machine)
    A = set()
    id2name = {}
    for e in root.iter("state"):
        if "name" in e.attrib:
            q = state_prefix(e.attrib["name"])
            id2name[e.attrib["id"]] = q
        else:
            q = state_prefix(e.attrib["id"])
            id2name[e.attrib["id"]] = q
        try:
            m.add_state(q)
        except AlreadyExistsState:
            pass
        if not e.find("initial") is None:
            m.set_initial_state(q)
        if not e.find("final") is None:
            A.add(q)
    m.set_aceptors(A)
    for e in root.iter("transition"):
        a = e.find("read").text
        if not a:
            a = "epsilon"
        else:
            try:
                m.add_symbol(a)
            except AlreadyExistsSymbol:
                pass
        q_i = id2name[e.find("from").text]
        q_f = id2name[e.find("to").text]
        if type_machine == 1:
            m.add_transition(q_i, a, q_f, force=True)
        else:
            m.add_transition(q_i, a, [q_f], force=True, update=True)

    return m



from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'maquinas',
  packages = ['maquinas','maquinas.regular','maquinas.contextfree','maquinas.parser','maquinas.contextsensitive','maquinas.recursivelyenumerable'],
  version = '0.3.1.6',
  license='GNU General Public License v3 or later',
  description = 'Formal languages and automata library',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Ivan Vladimir Meza Ruiz',
  author_email = 'ivanvladimir+maquinas@gmail.com',
  url = 'https://gitlab.com/ivanvladimir/maquinas',
  download_url = 'https://gitlab.com/ivanvladimir/maquinas/-/archive/v0.3.1.4/maquinas-v0.3.1.4.zip',
  keywords = ['regular languages', 'context free languages', 'context sensitive languages', 'recursively enumerable languages'],
  install_requires=[
          'graphviz',
          'ipywidgets',
          'IPython',
          'ordered_set',
          'Pillow',
          'TatSu'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Education',
    'Topic :: Education',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
  ],
)



# Código para Autómata Finito [Determinístic]
from maquinas.exceptions import *
from maquinas.regular.dfa import *
from maquinas.regular.RGParser import RGParser
import maquinas.parser.earley_parser as earley_parser
from ordered_set import OrderedSet
from IPython.core.display import display, HTML
import random
import itertools
import copy
import re

parser = RGParser()

# TODO: Check the left version and the extended


def _flat(vals, ini=[]):
    if isinstance(vals, tuple):
        ini1 = _flat(vals[0])
        ini2 = _flat(vals[1])
        return ini1 + ini2
    else:
        return [vals]


def _flat_tree(tree, acc=[]):
    if len(tree) == 2 and tree[-1][-1] == ";":
        acc = _flat_tree(tree[0])
        acc.append(tree[1])
        return acc
    return [tree]


class RegularGrammar:
    """Class for Context Free Grammar

    :param er_string: String with grammar
    :param S: Innitial symbol
    :param type: Type of grammar (default is 'right')
    """

    def __init__(self, er_string, S=None, type="right"):
        self.original = er_string
        self.Sigmariginal = er_string
        self.ast = parser.parse(er_string, rule_name="start")
        if len(self.ast) == 2:
            self.ast = _flat_tree(self.ast[0]) + [self.ast[1]]
        else:
            self.ast = _flat_tree(self.ast[0]) + [self.ast[1], self.ast[2]]
        self.V = set()
        self.sigma = set()
        self.P = {}
        replacements = {"epsilon": "ε"}
        for p in self.ast:
            head = p[0]
            if head not in self.sigma:
                self.V.add(head.replace('"', ""))
        for p in self.ast:
            body = _flat(p[2])
            body = [
                replacements.get(p.replace('"', ""), p.replace('"', "")) for p in body
            ]
            try:
                self.P[p[0]].append(body)
            except KeyError:
                self.P[p[0]] = [body]
            for val in body:
                if not val in self.V:
                    if not val == "ε":
                        self.sigma.add(val.replace('"', ""))
        if S:
            self.S = S
        else:
            self.S = self.ast[0][0]
        for alpha, betas in self.P.items():
            for beta in betas:
                if not alpha in self.V:
                    return "Error"
                if type == "right":
                    if (
                        not len(beta) == 2
                        and beta[0] in self.sigma
                        and beta[1] in self.V
                    ):
                        return "Error"
                elif type == "left":
                    if (
                        not len(beta) == 2
                        and beta[0] in self.sigma
                        and beta[1] in self.V
                    ):
                        return "Error"

        tokens = [t for t in self.V] + [t for t in self.sigma]
        tokens.sort(key=len, reverse=True)
        self.re_tokens = re.compile(
            "|".join(["".join([f"[{t_}]" for t_ in t]) + "\s*" for t in tokens])
        )

    def tokenize(self, string):
        return [t.strip() for t in self.re_tokens.findall(string)]

    def parse(self, string):
        """Parses a string with grammar

        :param string: String to be parsed

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tokens = self.tokenize(string)
        if len(tokens) == 0:
            return False, [], {}
        chart = earley_parser.parse(tokens, self.P, self.S, self.sigma)
        forest = earley_parser.extract_forest(self.S, chart, self.sigma, tokens)
        roots = earley_parser.verify_chart(chart, tokens, self.S)
        return roots, chart, forest

    def accepts(self, string):
        """Parses a string with grammar and determines if accepts it

        :param string: String to be parsed

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tokens = self.tokenize(string)
        if not len("".join(tokens)) == len(string):
            return False, [], {}
        chart = earley_parser.parse(tokens, self.P, self.S, self.sigma)
        roots = earley_parser.verify_chart(chart, tokens, self.S)
        return len(roots) > 0

    def _reduce_tree(self, tree):
        if len(tree) == 1:
            return tree
        head, children = tree
        subtree = []
        for child in children:
            child_ = self._reduce_tree(child)
            if child_[0]["partial"]:
                for grandchild in child_[1]:
                    subtree.append(grandchild)
            else:
                subtree.append(child_)
        return (head, subtree)

    def derivation(self, tree):
        """Iterator over derivation of a tree

        :param tree: Tree to retrieve derivation form

        :returns: Tuple with Roots of trres, chart parser and forest
        """
        tree = self._reduce_tree(tree)
        for i, nodes in enumerate(self.delta_analyse_tree(tree)):
            if i == 0:
                string = [("", "") for _ in range(nodes[0][1], nodes[0][2])]
            for node in nodes:
                ini, fin = node[1], node[2]
                if ini != fin:
                    pre, label = string[ini]
                    string[ini] = (pre, node[0])
                else:
                    if ini == len(string):
                        string.append(("", ""))
                    pre, label = string[ini]
                    string[ini] = (node[0], label)
            yield [l[0] + l[1] for l in string]

    def print_derivations(self, string, ini=0, fin=None):
        """Print derivations over trees

        :param string: String to be parsed
        :param ini: Initial tree to start printing from
        :param final: Final tree to print
        """
        roots, chart, forest = self.parse(string)
        if roots:
            trees = self.extract_trees(forest)
            for i, tree in enumerate(trees):
                if i <= ini:
                    continue
                if i > fin:
                    break
                print("Tree #", ini + i)
                for j, step in enumerate(self.derivation(tree)):
                    if not j:
                        print("".join(step), end="")
                    else:
                        print(" ⇒ ", "".join(step), end="\n ")
                print()

    def summary(self):
        """Generates summary of grammar"""
        info = [
            "No terminal : " + ", ".join(self.V),
            "Terminals   : " + ", ".join(self.sigma),
            "Start       : " + ", ".join(self.S),
            "Productions :\n"
            + "\n".join(
                f" {alpha} → {' | '.join(''.join(beta) for beta in betas)}"
                for alpha, betas in self.P.items()
            ),
        ]
        return "\n".join(info)

    def print_summary(self):
        """Prints summary of grammar"""
        return print(self.summary())

    def _delta_analyse_tree(self, tree):
        if len(tree) == 2:
            _, children = tree
            children_ = [t[0]["label"] for t in children]
            yield children_
            for child in children:
                yield from self._delta_analyse_tree(child)

    def delta_analyse_tree(self, tree):
        """Iteraror over analysis of a tree

        :param tree: Tree
        """
        tree = self._reduce_tree(tree)
        current, children = tree
        yield [current["label"]]
        children_ = [t[0]["label"] for t in children]
        yield children_
        for child in children:
            yield from self._delta_analyse_tree(child)

    def print_chart(self, string, chart, pointers=False):
        """Prints chart parser

        :param string: Original string fom parser
        :param chart: Chart parser
        :param pointers: Include pointers (default is no)
        """
        earley_parser.print_chart(string, chart, pointers=pointers)

    def chart2table(self, string, chart, pointers=False):
        """Creates HTML for chart parser

        :param string: Original string fom parser
        :param chart: Chart parser
        :param pointers: Include pointers (default is no)

        :returns: Display object for IPython"""
        earley_parser.chart2table(string, chart, pointers=pointers)

    def graph_forest(self, forest, **params):
        """Graph the forest structure

        :param forest: Forest structure
        :param params: Options to graph the tree

        :returns: Graphviz object"""
        return earley_parser.graph_forest(forest, **params)

    def extract_forest(self, chart):
        """Extract forest from chart object

        :param chart: Chart parser

        :returns: Forest data structure"""
        return earley_parser.extract_forest(self.S, chart, self.sigma)

    def extract_trees(self, forest, max_depth=None, max_ancesters=3):
        """Extract trees from forest structure

        :param forest: Forest data structure
        :param max_depth: Maximum depth to explore (default is None)
        :param max_ancester: Maximum number of the same ancester (default is 3)
        :returns: Trees data structure"""
        return earley_parser.extract_trees(
            forest, self.sigma, max_depth=max_depth, max_ancesters=max_ancesters
        )

    def graph_tree(self, tree, **params):
        """Graphs a tree

        :param tree: Tree data structure
        :param params: Options to graph the tree
        :returns: Graphviz objec"""
        return earley_parser.graph_tree(tree, **params)

    def graph_trees(self, trees, **params):
        """Graphs trees

        :param trees: Tree data structures
        :param params: Options to graph the tree
        :returns: Graphviz objec"""
        return earley_parser.graph_trees(trees, **params)

    def save_trees_img(self, trees, filename, **params):
        """Saves trees into an image

        :param trees: Tree data structures
        :param filename: Filename for the image file
        :param params: Options to graph the tree"""
        dot = self.graph_trees(trees)
        dot.render(filename, format="png", cleanup=True)

    def __str__(self):
        return self.original



# Código para Estado

from maquinas.exceptions import *
from maquinas.recursivelyenumerable.tm import TuringMachine


class Status:
    def __init__(self, machine, string="", state=None, memory=None):
        self.step = 0
        self.status = None
        self.position = 0
        self.machine = machine
        self.string = string
        if not machine is None:
            self.state = self.machine.delta_extended(None, "")
        else:
            self.state = None
        self.set_tape(string)
        self.memory = memory
        self.history = []
        self.positioncolor = "#b3e0ff"
        self.sucesscolor = "##00b300"
        self.failcolor = "#ff6666"

    def __str__(self):
        r = {
            "step": self.step,
            "tape": [_ for _ in self.tape],
            "pos": self.position,
            "state": self.state,
        }
        return str(r)

    def set_tape(self, string):
        if isinstance(self.machine, TuringMachine):
            self.tape = None
            for state in self.machine.delta_stepwise(string):
                self.state = state
                break
        else:
            self.tape = f"{string}"

    def get_symbol_tape(self):
        try:
            return self.tape[self.position]
        except IndexError:
            self.status = self.machine.accept(self.state)

    def get_memory(self):
        return self.memory

    def next(self, dir=-1):
        if not self.status is None:
            return None
        if self.step == len(self.history):
            prev = {
                "step": self.step,
                "tape": [_ for _ in self.tape] if self.tape else None,
                "pos": self.position,
                "state": self.state,
            }
            self.history.append(prev)
            try:
                self.machine.stepStatus(self)
            except DoesNotExistsSymbol:
                self.status = False
        else:
            step = self.history[self.step + 1]
            self.step = step["step"]
            self.tape = step["tape"]
            self.position = step["pos"]
            self.state = step["state"]
        if self.tape and self.position == len(self.tape):
            self.status = self.machine.acceptor(self.state)
        if not self.tape:
            if len(self.state) == 0:
                self.status = False
                self.state = self.history[-1]["state"]
            else:
                if self.machine.acceptor(self.state):
                    self.status = True

    def is_finish(self):
        if self.tape:
            return self.position == len(self.tape)
        else:
            return False

    def tape2html(self, ncols=20, symbols={}, states={}):
        if self.tape:
            tape_idx = list(enumerate(self.tape))
            tape = tape_idx[
                max(0, self.position - ncols) : min(
                    2 * ncols, self.position + ncols + 1
                )
            ]
            columns = []
            for pos, symbol in tape:
                symbol_ = symbols.get(symbol, symbol)
                if self.status is None:
                    if pos == self.position:
                        columns.append(
                            f"<td  bgcolor='{self.positioncolor}'>{symbol_}</td>"
                        )
                    else:
                        columns.append(f"<td>{symbol_}</td>")
                elif self.status:
                    columns.append(f"<td  bgcolor='{self.sucesscolor}'>{symbol_}</td>")
                else:
                    columns.append(f"<td  bgcolor='{self.failcolor}'>{symbol_}</td>")
            columns = "".join(columns)
            row = f"<tr>{columns}</tr>"
        if self.status is None:
            status = "Processing"
        elif self.status:
            status = "Accepted"
        else:
            status = "Rejected"

        state_ = self.machine._status(self.state, states=states, symbols=symbols)
        if self.tape:
            table = f"<strong>Estado actual:</strong> {state_}<br/><strong>Cadena:</strong> {status}<table>{row}</table>"
        else:
            table = f"<strong>Estado actual:</strong> {state_}<br/><strong>"
        return table
